int calc_wind_count(double point_x, double point_y, int vert_count, double *poly_vertices);
int * points_in_polygon(int *wind_counts, double *poly_vertices, int vert_count, double *points, int point_count);
